// You have generated a new plugin project without
// specifying the `--platforms` flag. A plugin project supports no platforms is generated.
// To add platforms, run `flutter create -t plugin --platforms <platforms> .` under the same
// directory. You can also find a detailed instruction on how to add platforms in the `pubspec.yaml` at https://flutter.dev/docs/development/packages-and-plugins/developing-packages#plugin-platforms.


import 'dart:math';
import 'dart:typed_data';

import 'package:flutter/services.dart';

class AlgoAngle {
  static const MethodChannel _channel = MethodChannel('algo_angle');
  final String version="Motion Algo Version V1.0.2";

  static Future<String?> get platformVersion async {
    final String? version = await _channel.invokeMethod('getPlatformVersion');
    return version;
  }

  String getVersion(){return version;}
  List<Vector3> unitVector = [Vector3(X:1, Y:0, Z:0),Vector3(X:0, Y:1, Z:0), Vector3(X:0, Y:0, Z:1)];

  TrainResult calc(final List<int> dataA,final List<int> dataB) {
    return _runThoughAlgo(dataA, dataB);
  }

  TrainResult _runThoughAlgo(final List<int> dataA, final List<int> dataB) {
    var arrA = Uint8List.fromList(dataA).buffer;
    ByteData _dataA = ByteData.view(arrA);
    var arrB = Uint8List.fromList(dataA).buffer;
    ByteData _dataB = ByteData.view(arrB);
    TrainDataPacket ap = _unpackBytes(_dataA);
    if (ap.errorCode != 0) {
      return TrainResult(errorCode: ap.errorCode);
    }
    TrainDataPacket bp = _unpackBytes(_dataB);
    if (bp.errorCode != 0) {
      return TrainResult(errorCode: bp.errorCode);
    }
    if ((ap.timestamp - bp.timestamp).abs() > 200) {
      return TrainResult(errorCode: 4);
    }
    List<double> qa = [];
    qa[0] = ap.queue[0];
    qa[1] = ap.queue[1];
    qa[2] = ap.queue[2];
    qa[3] = ap.queue[3];

    List<double> qb = [];
    qb[0] = bp.queue[0];
    qb[1] = bp.queue[1];
    qb[2] = bp.queue[2];
    qb[3] = bp.queue[3];
    if(_qIsNan(qa)){
      return TrainResult(errorCode: 5);
    }
    if(_qIsNan(qb)){
      return TrainResult(errorCode: 6);
    }
    List<double> qz=[0,0,0,1];
    qb=_qMultiple(qb,qz);

    List<Vector3> vaxisa=[];
    List<Vector3> vaxisb=[];
    for(int i=0;i<3;i++){
      vaxisa[i]=_vRotary(unitVector[i], qa);
      vaxisb[i]=_vRotary(unitVector[i], qb);
    }
    double angleA=0,angleB=0;
    Vector3 z=Vector3(X: 0,Y: 0,Z: -1);
    double gax=_vMultiple(z, vaxisa[0]);
    double gay=_vMultiple(z, vaxisa[1]);
    if((gay==0)&&(gay==0)){
      angleA=0;
    }else{
      angleA=atan2(gay, gax)*180/pi;
    }

    double gbx=_vMultiple(z,vaxisb[0]);
    double gby=_vMultiple(z,vaxisb[1]);

    if((gbx==0)&&(gby==0)){
      angleB=0;
    }else{
      angleB=atan2(gby, gbx)*180/pi;
    }

    double flexionextension=angleA-angleB;
    if(flexionextension>180){
      flexionextension=flexionextension-360;
    }else if(flexionextension<-180){
      flexionextension=flexionextension+360;
    }

    return TrainResult(flexionextension: flexionextension);
  }

  int toInt32(ByteData data, int idx) {
    // var arr = Uint8List.fromList(list).buffer;
    // ByteData data = ByteData.view(arr);
      print('to int32 ${data.getInt32(idx, Endian.little)}');
    return data.getInt32(idx, Endian.little);
  }

  double toFloat(ByteData data, int idx) {
    // var arr = Uint8List.fromList(list).buffer;
    // ByteData data = ByteData.view(arr);
        print('to float ${data.getFloat32(idx, Endian.little)}');
    return data.getFloat32(idx, Endian.little);
  }

  int toLong(ByteData data,int offset)
  {
      print('to long ${data.getInt64(offset,Endian.little)}');
     return data.getInt64(offset,Endian.little);
  }
   int _toUByte(final ByteData data, int offset) {
    print('to byte ${data.getInt32(offset,Endian.little) & 0xff}');
    return data.getInt32(offset,Endian.little) & 0xff;
  }

  TrainDataPacket _unpackBytes(final ByteData data) {
        TrainDataPacket P = TrainDataPacket(errorCode: 0);
        P.headSize = _toUByte(data, 0);
        P.dataSize = _toUByte(data, P.headSize + 1);
        if ((P.headSize + P.dataSize + 2) != data.lengthInBytes) {
            return TrainDataPacket(errorCode: 1);
        }
        P.headType = _toUByte(data, 1);
        P.dataSize = _toUByte(data, 2);
        if (P.headType != 0x88) {
           // Log.e(TAG, "Head type error");
            return TrainDataPacket(errorCode:2);
        }
        if (P.dataType != 0x02) {
            //Log.e(TAG, "data type error");
            return  TrainDataPacket(errorCode:3);
        }

        P.timestamp = toLong(data, 3);

        P.queue[0] = toFloat(data, P.headSize + 2 );
        P.queue[1] = toFloat(data, P.headSize + 2 + 1);
        P.queue[2] = toFloat(data, P.headSize + 2 + 2);
        P.queue[3] = toFloat(data, P.headSize + 2 + 3);

        return P;

    }

    

 bool _qIsNan(List<double> qu) {
    if (qu[0].isNaN) {
      return true;
    }
    if (qu[1].isNaN) {
      return true;
    }
    if (qu[2].isNaN) {
      return true;
    }
    if (qu[3].isNaN) {
      return true;
    }
    double normSquared =
        qu[0] * qu[0] + qu[1] * qu[1] + qu[2] * qu[2] + qu[3] * qu[3];
    if (normSquared == 0) {
      return true;
    } else {
      return false;
    }
  }
  double _vMultiple(Vector3 v0, Vector3 v1)
    {
        return v0.X * v1.X + v0.Y * v1.Y + v0.Z * v1.Z;
    }
  List<double> _qMultiple(List<double> qx,List<double>  qy) {
        List<double> qanswer = [0,0,0,0];
        double p0 = qx[0];
        double p1 = qx[1];
        double p2 = qx[2];
        double p3 = qx[3];
        double q0 = qy[0]; 
        double q1 = qy[1]; 
        double q2 = qy[2];
        double q3 = qy[3];

        qanswer[0] = p0 * q0 - p1 * q1 - p2 * q2 - p3 * q3;
        qanswer[1] = p0 * q1 + p1 * q0 + p2 * q3 - p3 * q2;
        qanswer[2] = p0 * q2 + p2 * q0 + p3 * q1 - p1 * q3;
        qanswer[3] = p0 * q3 + p3 * q0 + p1 * q2 - p2 * q1;
        return qanswer;
    }

     Vector3 _vNormalize(Vector3 V) {
        double norm = sqrt(V.X * V.X + V.Y * V.Y + V.Z * V.Z);
        return Vector3(X: V.X / norm, Y:V.Y / norm, Z:V.Z / norm);
    }

     Vector3 _vRotary(Vector3 V, List<double> Q) {

        double a = Q[0];
        double b = Q[1];
        double c = Q[2];
        double d = Q[3];
        double x = (1 - 2 * c * c - 2 * d * d) * V.X +
                (2 * b * c - 2 * a * d) * V.Y +
                (2 * a * c + 2 * b * d) * V.Z;

        double y = (2 * b * c + 2 * a * d) * V.X +
                (1 - 2 * b * b - 2 * d * d) * V.Y +
                (2 * c * d - 2 * a * b) * V.Z;

        double z = (2 * b * d - 2 * a * c) * V.X +
                (2 * a * b + 2 * c * d) * V.Y +
                (1 - 2 * b * b - 2 * c * c) * V.Z;

        Vector3 temp = Vector3(X:x, Y:y, Z:z);

        return _vNormalize(temp);
    }

  
}

class TrainDataPacket {
  TrainDataPacket(
      {this.errorCode = 0,
      this.headSize = 0,
      this.headType = 0,
      this.dataSize = 0,
      this.dataType = 0,
      this.timestamp = 0,
      this.queue = const [0, 0, 0, 0]});
  int errorCode;
  int headSize;
  int headType;
  int dataType;
  int timestamp;
  int dataSize;
  List<double> queue;
}

class TrainResult
{
  TrainResult({this.errorCode=0,this.flexionextension=0});
  int errorCode;
  double flexionextension;

  double getFlexiontension()
  {
    return flexionextension;
  }

}

class Vector3 {
  double X;
  double Y;
  double Z;

  Vector3({
    required this.X,
    required this.Y,
    required this.Z,
  });
}
